package aura.cm3d2.core;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;

class InstallState {
    final static class UninstallEntry {
        final String unknown_value;
        final String path;

        UninstallEntry(String unknown_value, String path) {
            this.unknown_value = unknown_value;
            this.path = path;
        }
    }
    final static class UpdateEntry {
        final String path;
        final int version;

        UpdateEntry(String path, int version) {
            this.path = path;
            this.version = version;
        }
    }

    final Path installPath;

    private static final String UNINST_NAME = "uninst.dat";
    private static final String UPDATE_NAME = "Update.lst";

    private final LockedFile uninstallLock;
    private final LockedFile updateLock;

    private final HashMap<String, UninstallEntry> uninstallEntries = new HashMap<>();
    private final HashMap<String, UpdateEntry> updateEntries = new HashMap<>();

    private InstallState(Path installPath, LockedFile uninstallLock, LockedFile updateLock) {
        this.installPath = installPath;
        this.uninstallLock = uninstallLock;
        this.updateLock = updateLock;
    }

    private void addUninstallEntry(UninstallEntry e) {
        uninstallEntries.put(Utils.normalizePath(e.path), e);
    }
    private void addUpdateEntry(UpdateEntry e) {
        updateEntries.put(Utils.normalizePath(e.path), e);
    }
    private void read() throws IOException {
        for(String uninstallLine : uninstallLock.getLines()) {
            String[] split = uninstallLine.split(",");
            addUninstallEntry(new UninstallEntry(split[0], split[1]));
        }
        for(String updateLine : updateLock.getLines()) {
            String[] split = updateLine.split(",");
            addUpdateEntry(new UpdateEntry(split[0], Integer.parseInt(split[1])));
        }
    }
    private void save() throws IOException {
        uninstallLock.writeList(uninstallEntries.values().toArray(new UninstallEntry[0]),
                x -> x.path, x -> x.unknown_value+","+x.path);
        updateLock.writeList(updateEntries.values().toArray(new UpdateEntry[0]),
                x -> x.path, x -> x.path+","+x.version);
    }

    int getVersion(String path) {
        path = Utils.normalizePath(path);
        if(!updateEntries.containsKey(path)) return -1;
        else return updateEntries.get(path).version;
    }

    void addToUninstallList(Path p) throws IOException {
        String path = p.toAbsolutePath().toString();
        if(!uninstallEntries.containsKey(Utils.normalizePath(path))) {
            addUninstallEntry(new UninstallEntry("0", path));
            save();
        }
    }
    void setVersion(String path, int version) throws IOException {
        addUpdateEntry(new UpdateEntry(path, version));
        save();
    }

    public void release() {
        uninstallLock.release();
        updateLock.release();
    }

    public static boolean exists(Path installPath) {
        return Files.exists(installPath.resolve(UNINST_NAME));
    }
    public static InstallState open(Path installPath) {
        // This code should only be run on Windows so Path should already be case insensitive
        LockedFile uninstLock = LockedFile.acquire(installPath.resolve(UNINST_NAME));

        Path updatePath = installPath.resolve(UPDATE_NAME);
        if(!Files.exists(updatePath)) try {
            Files.createFile(updatePath);
        } catch (IOException e) {
            throw new UpdaterException("Failed to create update state.", e);
        }
        LockedFile updateLock = LockedFile.acquire(updatePath);

        if(uninstLock != null && updateLock != null) {
            InstallState state = new InstallState(installPath, uninstLock, updateLock);
            try {
                state.read();
            } catch (IOException e) {
                throw new UpdaterException("Failed to parse installation state.", e);
            }
            try {
                state.addToUninstallList(updatePath);
            } catch (IOException e) {
                throw new UpdaterException("Failed to register update.lst to uninst.dat.", e);
            }
            return state;
        } else return null;
    }
}
