package aura.cm3d2.core;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.function.BiConsumer;
import java.util.zip.CRC32;

// TODO: Properly handle disk space checking on *nix

class TaskExecutor {
    private static final int GUI_UPDATE_PERIOD_MS = 100;
    private static final String PATH_INFO_FORMAT = "(\"%s\", %d, 0x%08x)";

    interface CompletionLambda {
        void onCompletion() throws IOException;
    }
    private static class FileCopy {
        public final UpdateFileList.Entry source;
        public final Path target;
        public final CompletionLambda onComplete;

        public FileCopy(UpdateFileList.Entry source, Path target, CompletionLambda onComplete) {
            this.source = source;
            this.target = target;
            this.onComplete = onComplete;
        }
    }

    public static class FileVerify {
        public final UpdatePackage.DisplayData display;
        public final UpdateFileList.Entry source;

        public FileVerify(UpdatePackage.DisplayData display, UpdateFileList.Entry source) {
            this.display = display;
            this.source = source;
        }
    }

    private ArrayList<FileCopy> copyList = new ArrayList<>();
    private ArrayList<FileVerify> verifyList = new ArrayList<>();
    private HashMap<Game, InstallState> installState = new HashMap<>();

    void copyEntry(UpdateFileList.Entry source, Path target, CompletionLambda onComplete) throws IOException {
        copyList.add(new FileCopy(source, target, onComplete));
    }
    void copyEntry(UpdateFileList.Entry source, Path target) throws IOException {
        copyEntry(source, target, () -> { });
    }

    void verifyFile(UpdatePackage.DisplayData display, UpdateFileList.Entry source) {
        verifyList.add(new FileVerify(display, source));
    }

    InstallState getInstallState(Game game) {
        if (!installState.containsKey(game)) installState.put(game, game.getInstallState());
        InstallState state = installState.get(game);
        if (state == null) throw new UpdaterException(Utils.getString("TaskList.installNoState"));
        return state;
    }

    private static class DriveInfo {
        long requiredSize = 0, maxPeakSize = 0;
    }
    private static class ProgressInfo {
        long lastUpdate = 0;
        long workCount  = 0;
        long totalSize  = 0;

        void doCheck(IProgressListener listener, long readCount) {
            workCount += readCount;

            long currentTime = System.currentTimeMillis();
            if(currentTime > lastUpdate + GUI_UPDATE_PERIOD_MS) {
                listener.checkCancel();
                listener.setWork(workCount, totalSize);
                lastUpdate = currentTime;
            }
        }
    }
    public void execute(IProgressListener listener, Settings settings) throws IOException {
        if (!verifyList.isEmpty()) {
            ProgressInfo progressInfo = new ProgressInfo();
            listener.setIndeterminate();
            listener.setCancelable(false);

            long totalSize = 0;
            for (FileVerify verify : verifyList) totalSize += verify.source.length;
            progressInfo.totalSize = totalSize;

            listener.setCancelable(true);

            Utils.debugPrint("Running extended verification...");

            long errorCount = 0;
            CRC32 crc = new CRC32();
            BiConsumer<ByteBuffer, Integer> callback = (buffer, readCount) -> {
                crc.update(buffer);
                progressInfo.doCheck(listener, readCount);
            };
            for (FileVerify verify : verifyList) {
                listener.setTask(Utils.format("TaskList.verifying", verify.source.outPath));

                if (!Files.exists(verify.source.source)) {
                    Utils.debugPrint(" - Verify error: Source file "+verify.source.source+" does not exist.");
                    progressInfo.workCount += verify.source.length;
                    errorCount += 1;
                    continue;
                }

                boolean isError = false;
                long actualSize = Files.size(verify.source.source);
                if (actualSize != verify.source.length) {
                    progressInfo.totalSize -= verify.source.length;
                    progressInfo.totalSize += actualSize;
                    isError = true;
                }

                crc.reset();
                Utils.read(verify.source.source, actualSize, callback);
                int actualCrc32 = (int) crc.getValue();
                if (actualCrc32 != verify.source.crc32) isError = true;

                if (isError) {
                    Utils.debugPrint(" - Verify error: Source file "+verify.source.source+" failed to verify.");
                    Utils.debugPrint("   - Source  : "+verify.display);
                    Utils.debugPrint("   - Expected: "+String.format(PATH_INFO_FORMAT,
                            verify.source.outPath, verify.source.length, verify.source.crc32));
                    Utils.debugPrint("   - Actual  : "+String.format(PATH_INFO_FORMAT,
                            verify.source.outPath, actualSize, actualCrc32));
                    errorCount += 1;
                }
            }
            if (errorCount > 0) throw new UpdaterException(Utils.format("TaskList.verifyFailed", errorCount));
        }

        if (!copyList.isEmpty()) {
            ProgressInfo progressInfo = new ProgressInfo();
            listener.setIndeterminate();
            listener.setCancelable(false);

            copyList.sort(Comparator.comparing(x -> -x.source.length));

            long totalSize = 0;
            HashMap<Path, DriveInfo> driveInfo = new HashMap<>();
            for (FileCopy copy : copyList) {
                Path root = copy.target.getRoot();
                if (!driveInfo.containsKey(root)) {
                    driveInfo.put(root, new DriveInfo());
                }
                DriveInfo drive = driveInfo.get(root);

                long existingSize = Files.exists(copy.target) ? Files.size(copy.target) : 0;
                totalSize += copy.source.length;
                drive.maxPeakSize = Long.max(drive.maxPeakSize, drive.requiredSize + copy.source.length);
                drive.requiredSize = drive.requiredSize - existingSize + copy.source.length;
            }
            progressInfo.totalSize = totalSize;

            if (!settings.disableDiskSpaceChecking) for (Map.Entry<Path, DriveInfo> drive : driveInfo.entrySet()) {
                long requiredSpace =
                        Long.max(Long.max(drive.getValue().requiredSize, drive.getValue().maxPeakSize), 0);
                Utils.checkAvailableDiskSpace(drive.getKey(), requiredSpace);
            }

            for (FileCopy copy : copyList) {
                if (!Files.exists(copy.source.source))
                    throw new UpdaterException(Utils.format("TaskList.missingSourceFile", copy.source.source));
                if (Files.size(copy.source.source) != copy.source.length)
                    throw new UpdaterException(Utils.format("TaskList.sizeMismatch", copy.target));
            }

            listener.setCancelable(true);

            CRC32 crc = settings.skipCrc32Verification ? null : new CRC32();
            BiConsumer<ByteBuffer, Integer> callback = (buffer, readCount) -> {
                if(crc != null) crc.update(buffer);
                progressInfo.doCheck(listener, readCount);
            };
            for(FileCopy copy : copyList) {
                listener.setTask(Utils.format("TaskList.copying", copy.source.outPath));

                Path tempPath = copy.target.getParent().resolve(copy.target.getFileName().toString()+".tmp");
                Files.createDirectories(copy.target.getParent());

                if (crc != null) crc.reset();

                try {
                    listener.checkCancel();
                    Utils.copy(copy.source.source, tempPath, copy.source.length, callback);
                    if(crc != null && (int) crc.getValue() != copy.source.crc32)
                        throw new UpdaterException(Utils.format("UpdatePackage.crcMismatch", copy.target));
                    Files.move(tempPath, copy.target, StandardCopyOption.REPLACE_EXISTING);
                    copy.onComplete.onCompletion();
                } finally {
                    Files.deleteIfExists(tempPath);
                }
            }

            for (InstallState state : installState.values()) {
                state.release();
            }
        }

        listener.checkCancel();
        listener.done();
    }
}

public class TaskList {
    interface InitFunction {
        void onInit(IProgressListener listener, TaskExecutor executor, Settings settings) throws IOException;
    }
    private ArrayList<InitFunction> initList = new ArrayList<>();

    void onInit(InitFunction init) {
        initList.add(init);
    }

    public void execute(IProgressListener listener, Settings settings) throws IOException {
        listener.setIndeterminate();
        listener.setCancelable(false);

        TaskExecutor executor = new TaskExecutor();
        for (InitFunction init : initList) init.onInit(listener, executor, settings);
        executor.execute(listener, settings);
    }

    public static void enableDebugPrint() {
        Utils.enableDebugPrint();
    }
}
