package aura.cm3d2.core;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;

public final class UpdateFileList {
    public static final class Entry {
        public final Path source;
        public final String outPath;
        public final String canonicalPath;
        public final long length;
        public final int crc32;
        public final int version;

        private Entry(Path source, String outPath, String canonicalPath, long length, int crc32, int version) {
            this.source = source;
            this.outPath = outPath;
            this.canonicalPath = canonicalPath;
            this.length = length;
            this.crc32 = crc32;
            this.version = version;
        }

        public boolean isSameFile(Entry other) {
            return canonicalPath.equals(other.canonicalPath) &&
                    length == other.length && crc32 == other.crc32 && version == other.version;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            Entry that = (Entry) o;
            return length == that.length &&
                    crc32 == that.crc32 &&
                    version == that.version &&
                    Objects.equals(outPath, that.outPath);
        }

        @Override
        public int hashCode() {
            return Objects.hash(outPath, length, crc32, version);
        }
    }

    private HashMap<String, Entry> entries = new HashMap<>();

    private void addEntry(Entry entry) {
        if(entries.containsKey(entry.canonicalPath)) {
            Entry oldEntry = entries.get(entry.canonicalPath);
            if(oldEntry.version == entry.version) {
                if(oldEntry.length != entry.length ||
                   oldEntry.crc32 != entry.crc32)
                    throw new UpdaterException("Two differing versions of file "+oldEntry.outPath+" were found!");
                entries.put(entry.canonicalPath, entry);
            } else if(entry.version > oldEntry.version) entries.put(entry.canonicalPath, entry);
        } else {
            entries.put(entry.canonicalPath, entry);
        }
    }

    private void load(Path path) throws IOException {
        Path rootPath = path.getParent();
        for(String line : Utils.readFileAsLines(path))  try {
            String[] data = line.trim().split(",");

            Path source = data[0].equals("share") ? Utils.caseInsensitiveResolve(rootPath, data[1])
                                                  : Utils.caseInsensitiveResolve(rootPath, "data", data[2]);
            String outPath = Utils.normalizePathOutput(data[2]);
            String canonicalPath = Utils.normalizePath(data[2]);
            long length = Long.parseLong(data[3]);
            int crc32 = (int) Long.parseLong(data[4], 16);
            int version = Integer.parseInt(data[5]);
            addEntry(new Entry(source, outPath, canonicalPath, length, crc32, version));
        } catch(ArrayIndexOutOfBoundsException | NumberFormatException e) {
            throw new UpdaterException("Failed to parse package update list.", e);
        }
    }

    public static UpdateFileList loadUpdateFile(Path path) throws IOException {
        UpdateFileList list = new UpdateFileList();
        list.load(path);
        return list;
    }

    public void writeUpdateFile(Path path) throws IOException {
        Utils.writeList(path, getEntries(), x -> x.outPath, entry ->
            "0,0,"+entry.outPath+","+ entry.length+","+String.format("%08X", entry.crc32)+","+entry.version);
    }

    void copyMissingFrom(UpdateFileList other) {
        Entry[] entries = getEntries();
        for (Entry entry : entries)
            if (!Files.exists(entry.source) && other.entries.containsKey(entry.canonicalPath)) {
                Entry otherEntry = other.entries.get(entry.canonicalPath);
                if (Files.exists(otherEntry.source) && otherEntry.isSameFile(entry)) {
                    Utils.debugPrint("WARNING: Copying missing "+entry.outPath+" from non-CBL version.");
                    this.entries.put(entry.canonicalPath, otherEntry);
                }
            }
    }

    public boolean isEmpty() {
        return entries.isEmpty();
    }

    public Entry[] getEntries() {
        return entries.values().toArray(new Entry[0]);
    }
    public Entry[] getSortedEntries() {
        Entry[] entries = getEntries();
        Arrays.sort(entries, Comparator.comparing(pkg -> pkg.canonicalPath));
        return entries;
    }

    public UpdateFileList merge(UpdateFileList other) {
        UpdateFileList list = new UpdateFileList();
        for(Entry e : entries.values()) list.addEntry(e);
        for(Entry e : other.entries.values()) list.addEntry(e);
        return list;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        UpdateFileList that = (UpdateFileList) o;
        return Objects.equals(entries, that.entries);
    }

    @Override
    public int hashCode() {
        return Objects.hash(entries);
    }
}
