package aura.cm3d2.core;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.Collator;
import java.util.*;

public class UpdatePackage {
    // This is used by this program to identify its own installers, thus is not I18N'd.
    private final static String CBI_MARKER_STR = "  Repacked by CM3D2 Batch Installer\t";
    private final static String CUM_MARKER_STR = "This package was created with CUM";

    private final static String UPDATER_PRODUCT_NAME = "Reactorアップデート";

    private final static Collator COLLATOR;
    static {
        COLLATOR = Collator.getInstance(Locale.JAPANESE);
        COLLATOR.setStrength(Collator.PRIMARY);
        COLLATOR.setDecomposition(Collator.FULL_DECOMPOSITION);
    }

    public enum DisplayTag {
        PLUGIN(Utils.getString("UpdatePackage.DisplayTag.Plugin")),
        UPDATER(Utils.getString("UpdatePackage.DisplayTag.Updater")),
        REPACK(Utils.getString("UpdatePackage.DisplayTag.Repack")),
        UNKNOWN_REPACK(Utils.getString("UpdatePackage.DisplayTag.UnknownRepack")),
        MERGED_PACKAGE(Utils.getString("UpdatePackage.DisplayTag.MergedPackage")),
        UNKNOWN(Utils.getString("UpdatePackage.DisplayTag.Unknown"));

        private final String display;

        DisplayTag(String display) {
            this.display = display;
        }

        @Override
        public String toString() {
            return display;
        }
    }

    public static class DisplayData {
        public final DisplayTag tag;
        public final String name;

        public DisplayData(DisplayTag tag, String name) {
            this.tag = tag;
            this.name = name;
        }

        public String toHtmlString() {
            return "(<i>"+tag+"</i>) "+name;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            DisplayData that = (DisplayData) o;
            return tag == that.tag && Objects.equals(name, that.name);
        }

        @Override
        public int hashCode() {
            return Objects.hash(tag, name);
        }

        @Override
        public String toString() {
            return "("+tag+") "+name;
        }
    }
    private final static DisplayData MERGED_CBI_DISPLAY =
            new DisplayData(DisplayTag.MERGED_PACKAGE, Utils.getString("UpdatePackage.merged.CBI"));
    private final static DisplayData MERGED_CUM_DISPLAY =
            new DisplayData(DisplayTag.MERGED_PACKAGE, Utils.getString("UpdatePackage.merged.CUM"));

    public final Game game;
    public final DisplayData display;

    public final UpdaterBinary binary;
    public final String readme;
    public final UpdateMetadata metadata;

    private ArrayList<String> extraReadmes;

    public final UpdateFileList updateX86;
    public final UpdateFileList updateX64;

    public UpdatePackage(Game game, DisplayData display, UpdaterBinary binary, String readme,
                         UpdateMetadata metadata, ArrayList<String> extraReadmes,
                         UpdateFileList updateX86, UpdateFileList updateX64) {
        this.game = game;
        this.display = display;
        this.binary = binary;
        this.readme = readme;
        this.metadata = metadata;
        this.extraReadmes = extraReadmes;
        this.updateX86 = updateX86;
        this.updateX64 = updateX64;
    }

    private static UpdateFileList readLstIfExists(Path path, UpdateFileList def) throws IOException {
        if(!Files.exists(path)) return def;
        else return UpdateFileList.loadUpdateFile(path);
    }

    private static String stripOuterQuotes(String str) {
        if(str.startsWith("『") && str.endsWith("』")) return str.substring(1, str.length() - 1);
        return str;
    }
    private static DisplayData getDisplay(Path sourceRoot, Game game, String readme) {
        String[] split = readme.split("[\r\n]+");

        // Merged packages from various sources.
        if (readme.contains(CBI_MARKER_STR)) return MERGED_CBI_DISPLAY;
        if (readme.contains(CUM_MARKER_STR)) return MERGED_CUM_DISPLAY;

        // Plugin-like packages
        block: {
            if (split.length < 4) break block;
            if (!split[0].matches("^=+$")) break block;
            if (!split[3].matches("^=+$")) break block;
            if (Arrays.stream(game.displayProductAlias).noneMatch(displayProduct -> {
                String updaterText = "『"+displayProduct+"』専用プラグイン";
                return COLLATOR.equals(split[1].trim(), updaterText);
            })) break block;
            return new DisplayData(DisplayTag.PLUGIN, stripOuterQuotes(split[2].trim()));
        }

        // Updater-like packages
        block: {
            if (split.length < 3) break block;
            if (!split[0].matches("^=+$")) break block;
            if (!split[2].matches("^=+$")) break block;

            String[] split2 = split[1].trim().split("』", 2);
            if (!split2[0].startsWith("『")) break block;
            String gameString = split2[0].substring(1).trim();
            if (Arrays.stream(game.displayProductAlias).noneMatch(displayProduct ->
                    COLLATOR.equals(gameString, displayProduct)
            )) break block;

            String updateText = split2[1].trim();
            if (updateText.startsWith("アップデータ")) updateText = updateText.substring(6).trim();

            return new DisplayData(DisplayTag.UPDATER, updateText);
        }

        // Unknown package types
        String rootStr = sourceRoot == null ? "<no path>" : sourceRoot.toString();

        // Repacks
        if (readme.toLowerCase(Locale.JAPANESE).contains("repack"))
            return new DisplayData(DisplayTag.UNKNOWN_REPACK, rootStr);

        return new DisplayData(DisplayTag.UNKNOWN, rootStr);
    }

    public static UpdatePackage loadPackage(Path path) {
        UpdaterBinary binary;
        try {
            binary = UpdaterBinary.read(Utils.caseInsensitiveResolve(path, "update.exe"));
        } catch(Exception e) {
            throw new UpdaterException("Failed to read updater binary: "+e.getMessage(), e);
        }
        if (!binary.productName.contains(UPDATER_PRODUCT_NAME)) {
            throw new UpdaterException("Unknown updater type: "+binary.productName);
        }

        UpdateMetadata metadata;
        try {
            metadata = UpdateMetadata.read(Utils.caseInsensitiveResolve(path, "update.ini"));
        } catch(Exception e) {
            throw new UpdaterException("Failed to read updater metadata: "+e.getMessage(), e);
        }

        String readmeData;
        try {
            readmeData = Utils.readFileAsString(Utils.caseInsensitiveResolve(path, "readme.txt"));
        } catch(Exception e) {
            throw new UpdaterException("Failed to read updater readme: "+e.getMessage(), e);
        }

        Game game = Game.checkGame(metadata);
        if (game == null) throw new UpdaterException("Updater is for an unknown game.");

        DisplayData display = getDisplay(path, game, readmeData);

        UpdateFileList updateShare, updateX86, updateX64;
        try {
            updateShare = readLstIfExists(Utils.caseInsensitiveResolve(path, "update.lst"), null);
            updateX86   = readLstIfExists(Utils.caseInsensitiveResolve(path, "update_X86.lst"), updateShare);
            updateX64   = readLstIfExists(Utils.caseInsensitiveResolve(path, "update_X64.lst"), updateShare);
        } catch (Exception e) {
            throw new UpdaterException("Failed to read updater file lists: "+e.getMessage(), e);
        }

        if (updateX86 == null) updateX86 = new UpdateFileList();
        if (updateX64 == null) updateX64 = new UpdateFileList();

        UpdatePackage tempPkg = new UpdatePackage(
                game, display, binary, readmeData, metadata, new ArrayList<>(), updateX86, updateX64);
        display = Patches.patchDisplayData(tempPkg);
        return new UpdatePackage(
            game, display, binary, readmeData, metadata, new ArrayList<>(), updateX86, updateX64);
    }

    public void writePackageToPath(TaskList tasks, Path out) {
        tasks.onInit((listener, executor, settings) -> {
            listener.setTask(Utils.getString("UpdatePackage.writingUpdater"));
            Files.createDirectories(out);
            Files.createDirectories(out.resolve("readme"));

            Files.copy(binary.source, out.resolve("update.exe"));
            Utils.writeFile(out.resolve("readme.txt"), readme);
            metadata.write(out.resolve("update.ini"));

            int fileId = 0;
            for (String readme : extraReadmes) {
                String fileName = String.format("readme_%05d.txt", fileId++);
                Utils.writeFile(out.resolve("readme").resolve(fileName), readme);
            }

            UpdateFileList entries = updateX86.merge(updateX64);
            if(!settings.dontMergeUpdateArchs) {
                entries.writeUpdateFile(out.resolve("update.lst"));
            } else if(updateX86.equals(updateX64)) {
                updateX86.writeUpdateFile(out.resolve("update.lst"));
            } else {
                updateX86.writeUpdateFile(out.resolve("update_x86.lst"));
                updateX64.writeUpdateFile(out.resolve("update_x64.lst"));
            }

            for (UpdateFileList.Entry entry : entries.getEntries()) {
                executor.copyEntry(entry, out.resolve("data").resolve(entry.outPath));
            }
        });
    }

    public void installPackage(TaskList tasks) {
        tasks.onInit((listener, executor, settings) -> {
            InstallState state = executor.getInstallState(game);
            listener.setTask(Utils.getString("UpdatePackage.preparingToUpdate"));

            UpdateFileList entries;
            if(!settings.dontMergeUpdateArchs) entries = updateX86.merge(updateX64);
            else entries = Utils.isWin64() ? updateX64 : updateX86;

            for (UpdateFileList.Entry entry : entries.getEntries()) {
                if(state.getVersion(entry.outPath) >= entry.version) continue;
                Path target = Utils.caseInsensitiveResolve(state.installPath, entry.outPath);
                executor.copyEntry(entry, target, () -> {
                    if(settings.registerWithUninst) state.addToUninstallList(target);
                    state.setVersion(entry.outPath, entry.version);
                });
            }
        });
    }

    public void verifyPackage(TaskList tasks) {
        tasks.onInit((listener, executor, settings) -> {
            UpdateFileList entries = updateX86.merge(updateX64);
            for (UpdateFileList.Entry entry : entries.getEntries()) {
                executor.verifyFile(display, entry);
            }
        });
    }

    private static void pushUpdateFileList(MessageDigest digest, UpdateFileList list) {
        for (UpdateFileList.Entry entry : list.getSortedEntries()) {
            String canonical = entry.canonicalPath+"\n"+entry.length+"\n"+entry.crc32+"\n\n";
            digest.update((canonical).getBytes(StandardCharsets.UTF_8));
        }
    }
    public String fileEntryHash() {
        try {
            MessageDigest digest = MessageDigest.getInstance("SHA-256");
            digest.update("[x86]\n".getBytes(StandardCharsets.UTF_8));
            pushUpdateFileList(digest, updateX86);
            digest.update("[x64]\n".getBytes(StandardCharsets.UTF_8));
            pushUpdateFileList(digest, updateX64);
            return Base64.getEncoder().encodeToString(digest.digest());
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }

    private void copyMissingFrom(UpdatePackage pkg) {
        updateX86.copyMissingFrom(pkg.updateX86);
        updateX64.copyMissingFrom(pkg.updateX64);
    }

    private static String buildReadme(Game game, UpdatePackage[] packages) {
        StringWriter strOut = new StringWriter();
        PrintWriter out = new PrintWriter(strOut);

        String[] sorted = Arrays.stream(packages).map(pkg -> pkg.display.toString()).toArray(String[]::new);
        Arrays.sort(sorted, Comparator.naturalOrder());
        for(String pkg : sorted) out.println(" - "+pkg);

        return String.format(Utils.getString("UpdatePackage.mergeReadme"),
                game.mergeDisplayProduct, packages.length, strOut.toString().replaceAll("\n$", ""), CBI_MARKER_STR);
    }
    public static UpdatePackage mergePackagesForGame(Game game, UpdatePackage... packages) {
        UpdaterBinary binary = packages[0].binary;
        for(UpdatePackage pkg : packages) if(pkg.binary.productVersion > binary.productVersion) binary = pkg.binary;

        String readme = buildReadme(game, packages);

        for(UpdatePackage pkg : packages)
            if(!pkg.metadata.isGame(game))
                throw new RuntimeException("Package registry information does not match provided update metadata!");

        ArrayList<String> extraReadmes = new ArrayList<>();
        for(UpdatePackage pkg : packages) extraReadmes.add(pkg.readme);

        UpdateFileList updateX86 = new UpdateFileList();
        UpdateFileList updateX64 = new UpdateFileList();
        for(UpdatePackage pkg : packages) {
            updateX86 = updateX86.merge(pkg.updateX86);
            updateX64 = updateX64.merge(pkg.updateX64);
        }
        if (updateX86.equals(updateX64)) updateX86 = updateX64;

        return new UpdatePackage(
                game, MERGED_CBI_DISPLAY, binary, readme, game.getMergedMetadata(),
                extraReadmes, updateX86, updateX64);
    }
    public static HashMap<Game, UpdatePackage> mergePackages(UpdatePackage... packages) {
        HashMap<Game, ArrayList<UpdatePackage>> packageList = new HashMap<>();
        for (Game game : Game.values()) packageList.put(game, new ArrayList<>());
        for (UpdatePackage pkg : packages) packageList.get(pkg.game).add(pkg);

        HashMap<Game, UpdatePackage> mergedPackages = new HashMap<>();
        for (Game game : Game.values()) {
            if (packageList.get(game).isEmpty()) continue;
            mergedPackages.put(game, mergePackagesForGame(game, packageList.get(game).toArray(new UpdatePackage[0])));
        }
        if (mergedPackages.containsKey(Game.CM3D2) && mergedPackages.containsKey(Game.CM3D2ChuB))
            mergedPackages.get(Game.CM3D2ChuB).copyMissingFrom(mergedPackages.get(Game.CM3D2));
        if (mergedPackages.containsKey(Game.COM3D2) && mergedPackages.containsKey(Game.COM3D2ChuB))
            mergedPackages.get(Game.COM3D2ChuB).copyMissingFrom(mergedPackages.get(Game.COM3D2));
        return mergedPackages;
    }

    public static TaskList installPackages(UpdatePackage... rawPackages) {
        TaskList tasks = new TaskList();
        HashMap<Game, UpdatePackage> packages = mergePackages(rawPackages);
        for (Game game : packages.keySet()) {
            if (game.isInstalled()) {
                packages.get(game).installPackage(tasks);
            }
        }
        return tasks;
    }
    public static TaskList mergePackages(Path outPath, UpdatePackage... rawPackages) {
        TaskList tasks = new TaskList();
        HashMap<Game, UpdatePackage> packages = mergePackages(rawPackages);
        if (packages.size() == 0) {
            packages.values().stream().findAny().get().writePackageToPath(tasks, outPath);
        } else {
            for (Game game : packages.keySet()) {
                Path gamePath = outPath.resolve(game.mergeDisplayProduct);
                packages.get(game).writePackageToPath(tasks, gamePath);
            }
        }
        return tasks;
    }
}
