package aura.cm3d2.core;

import aura.cm3d2.core.win32.Kernel32Util;
import aura.cm3d2.core.win32.SYSTEM_INFO;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.*;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.stream.Stream;

class Utils {
    private static final ResourceBundle TEXT = ResourceBundle.getBundle("aura/cm3d2/core/text");
    static final String SJIS = "Shift_JIS";

    static String getString(String s) {
        return TEXT.getString(s);
    }
    static String format(String s, Object... format) {
        return String.format(TEXT.getString(s), format);
    }

    // Modified from https://stackoverflow.com/a/24805871
    private static final String[] UNIT_LIST = new String[]{"KiB", "MiB", "GiB", "TiB", "PiB", "EiB"};
    private static String formatFileSize(long value) {
        int bitSize = (63 - Long.numberOfLeadingZeros(value));
        int siScale = (bitSize / 10) - 1;
        if(siScale < 0) siScale = 0;
        return String.format("%.2f %s", (double) value / (1L << ((siScale + 1) * 10)), UNIT_LIST[siScale]);
    }

    static void checkAvailableDiskSpace(Path path, long required) throws IOException {
        long available = Files.getFileStore(path).getUsableSpace();
        if(available < required)
            throw new UpdaterException(format("Utils.insufficientDiskSpace",
                                       path, formatFileSize(available), formatFileSize(required)));
    }

    private static Path caseInsensitiveResolveComponent(Path path, String component) {
        Path out = path.resolve(component);
        if(!Files.exists(out))
            try {
                try (Stream<Path> paths = Files.list(path)) {
                    for (Iterator<Path> it = paths.iterator(); it.hasNext(); ) {
                        Path p = it.next();
                        if(p.getFileName().toString().equalsIgnoreCase(component))
                            return p;
                    }
                }
            } catch (IOException e) {
                // fall through
            }
        return out;
    }
    static Path caseInsensitiveResolve(Path root, String... paths) {
        for(String path : paths)
            for(String component : path.replace("\\", "/").split("/"))
                switch(component) {
                    case ".":
                        break;
                    case "..":
                        root = root.getParent();
                        break;
                    default:
                        root = caseInsensitiveResolveComponent(root, component);
                }
        return root;
    }

    private static void transfer(Path from, WritableByteChannel targetChannel, long toTransfer,
                                 BiConsumer<ByteBuffer, Integer> callback) throws IOException {
        FileChannel fileChannel = FileChannel.open(from, StandardOpenOption.READ);
        try {
            WritableByteChannel targetWrapper = new WritableByteChannel() {
                @Override
                public int write(ByteBuffer src) throws IOException {
                    int n;
                    int position = src.position();
                    if ((n = targetChannel.write(src)) > 0) {
                        int newPosition = src.position();
                        src.position(position);
                        ByteBuffer view = src.slice();
                        view.limit(n);
                        callback.accept(view, n);
                        src.position(newPosition);
                    }
                    return n;
                }

                @Override
                public boolean isOpen() {
                    return targetChannel.isOpen();
                }

                @Override
                public void close() throws IOException {
                    targetChannel.close();
                }
            };

            long position = 0;
            int zeroTxCount = 0;
            while (toTransfer > 0) {
                long transferred = fileChannel.transferTo(position, toTransfer, targetWrapper);
                if (transferred == 0) zeroTxCount++;
                if (zeroTxCount > 100) throw new UpdaterException("Transfer stuck?");
                toTransfer -= transferred;
                position += transferred;
            }
        } finally {
            try {
                fileChannel.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }
    private static final WritableByteChannel NULL_CHANNEL = new WritableByteChannel() {
        @Override
        public int write(ByteBuffer src) {
            return src.remaining();
        }

        @Override
        public boolean isOpen() {
            return true;
        }

        @Override
        public void close() {

        }
    };
    static void read(Path from, long toTransfer, BiConsumer<ByteBuffer, Integer> callback) throws IOException {
        transfer(from, NULL_CHANNEL, toTransfer, callback);
    }
    static void copy(Path from, Path to, long toTransfer,
                     BiConsumer<ByteBuffer, Integer> callback) throws IOException {
        FileChannel outputChannel = FileChannel.open(to,
                StandardOpenOption.CREATE, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.WRITE);
        try {
            transfer(from, outputChannel, toTransfer, callback);
        } finally {
            try {
                outputChannel.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    static String normalizeEndings(String s) {
        return s.replace("\r\n", "\n").replace("\n", "\r\n");
    }
    static void writeFile(Path path, String s) throws IOException {
        Files.write(path, fromString(normalizeEndings(s)));
    }
    static byte[] fromString(String data) {
        try {
            return data.getBytes(SJIS);
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
    }
    static String toString(byte[] data) {
        try {
            return new String(data, SJIS);
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
    }
    static String readFileAsString(Path path) throws IOException {
        return toString(Files.readAllBytes(path));
    }

    static String[] splitLines(String s) {
        ArrayList<String> out = new ArrayList<>();
        for(String line : s.split("[\r\n]")) {
            String trim = line.trim();
            if(!trim.isEmpty()) out.add(trim);
        }
        return out.toArray(new String[0]);
    }
    static String[] readFileAsLines(Path path) throws IOException {
        return splitLines(readFileAsString(path));
    }

    static String normalizePath(String s) {
        return normalizePathOutput(s).toLowerCase(Locale.JAPANESE);
    }
    static String normalizePathOutput(String s) {
        return s.replace("\\", "/").replaceAll("/+", "/").replaceAll("^/+", "").replaceAll("/+$", "");
    }

    private static <T, U extends Comparable<? super U>> T[] sortBy(T[] array, Function<T, U> comparator) {
        T[] arr = Arrays.copyOf(array, array.length);
        Arrays.sort(arr, Comparator.comparing(comparator));
        return arr;
    }

    static <T, U extends Comparable<? super U>> String writeList(T[] array,
                                                                 Function<T, U> sortFn,
                                                                 Function<T, String> fn) {
        StringWriter strOut = new StringWriter();
        PrintWriter out = new PrintWriter(strOut);
        for(T entry : Utils.sortBy(array, sortFn)) out.println(fn.apply(entry));
        return strOut.toString();
    }
    static <T, U extends Comparable<? super U>> void writeList(Path path, T[] array,
                                                               Function<T, U> sortFn,
                                                               Function<T, String> fn) throws IOException {
        Utils.writeFile(path, writeList(array, sortFn, fn));
    }

    static boolean isWin64() {
        return Kernel32Util.getSystemInfo().processorArchitecture.pi.wProcessorArchitecture.intValue() ==
                SYSTEM_INFO.PROCESSOR_ARCHITECTURE_AMD64;
    }

    // ENABLE_DEBUG in case someone wants to use this as a library
    private static boolean ENABLE_DEBUG = false;
    static void enableDebugPrint() {
        ENABLE_DEBUG = true;
    }
    static void debugPrint(String string) {
        if (ENABLE_DEBUG) {
            System.out.println(string);
        }
    }
}
