/* Copyright (c) 2010 Daniel Doubrovkine, All Rights Reserved
 *
 * The contents of this file is dual-licensed under 2 
 * alternative Open Source/Free licenses: LGPL 2.1 or later and 
 * Apache License 2.0. (starting with JNA version 4.0.0).
 * 
 * You can freely decide which license you want to apply to 
 * the project.
 * 
 * You may obtain a copy of the LGPL License at:
 * 
 * http://www.gnu.org/licenses/licenses.html
 * 
 * A copy is also included in the downloadable source code package
 * containing JNA, in file "LGPL2.1".
 * 
 * You may obtain a copy of the Apache License at:
 * 
 * http://www.apache.org/licenses/
 * 
 * A copy is also included in the downloadable source code package
 * containing JNA, in file "AL2.0".
 */

package aura.cm3d2.core.win32;

import com.sun.jna.Native;
import com.sun.jna.ptr.IntByReference;
import com.sun.jna.win32.W32APIOptions;

/**
 * Advapi32 utility API.
 *
 * @author dblock[at]dblock.org
 */
public abstract class Advapi32Util {
    private static Advapi32 INSTANCE = null;
    private static synchronized Advapi32 getInstance() {
        if(INSTANCE == null)
            INSTANCE = Native.loadLibrary("Advapi32", Advapi32.class, W32APIOptions.DEFAULT_OPTIONS);
        return INSTANCE;
    }

    private static final int READ_CONTROL = 0x00020000;
    private static final int STANDARD_RIGHTS_READ = READ_CONTROL;
    private static final int KEY_QUERY_VALUE = 0x0001;
    private static final int KEY_ENUMERATE_SUB_KEYS = 0x0008;
    private static final int KEY_NOTIFY = 0x0010;
    private static final int SYNCHRONIZE = 0x00100000;

    private static final int KEY_READ =
            STANDARD_RIGHTS_READ | KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS | KEY_NOTIFY & (~SYNCHRONIZE);

    private static final int ERROR_SUCCESS = 0;
    private static final int ERROR_INSUFFICIENT_BUFFER = 122;

    private static final int REG_SZ = 1;
    private static final int REG_EXPAND_SZ = 2;

    /**
     * Get a registry REG_SZ value.
     *
     * @param root  Root key.
     * @param key   Registry path.
     * @param value Name of the value to retrieve.
     * @return String value.
     */
    public static String registryGetStringValue(WinReg.HKEY root, String key, String value) {
        Advapi32 advapi32 = getInstance();
        WinReg.HKEYByReference phkKey = new WinReg.HKEYByReference();
        int rc = advapi32.RegOpenKeyEx(root, key, 0, KEY_READ, phkKey);
        if (rc != ERROR_SUCCESS) throw new RuntimeException("Status: "+rc);
        try {
            return registryGetStringValue(phkKey.getValue(), value);
        } finally {
            rc = advapi32.RegCloseKey(phkKey.getValue());
            if (rc != ERROR_SUCCESS) throw new RuntimeException("Status: "+rc);
        }
    }

    /**
     * Get a registry REG_SZ value.
     *
     * @param hKey  Parent Key.
     * @param value Name of the value to retrieve.
     * @return String value.
     */
	private static String registryGetStringValue(WinReg.HKEY hKey, String value) {
        Advapi32 advapi32 = getInstance();
		IntByReference lpcbData = new IntByReference();
		IntByReference lpType = new IntByReference();
		int rc = advapi32.RegQueryValueEx(hKey, value, 0, lpType, (char[]) null, lpcbData);
		if (rc != ERROR_SUCCESS && rc != ERROR_INSUFFICIENT_BUFFER) throw new RuntimeException("Status: "+rc);
		if (lpType.getValue() != REG_SZ && lpType.getValue() != REG_EXPAND_SZ)
			throw new RuntimeException("Unexpected registry type " + lpType.getValue() +
                                       ", expected REG_SZ or REG_EXPAND_SZ");
		char[] data = new char[lpcbData.getValue()];
		rc = advapi32.RegQueryValueEx(hKey, value, 0, lpType, data, lpcbData);
		if (rc != ERROR_SUCCESS && rc != ERROR_INSUFFICIENT_BUFFER) throw new RuntimeException("Status: "+rc);
		return Native.toString(data);
	}
}
