package aura.cm3d2.ui;

import aura.cm3d2.core.*;
import aura.cm3d2.core.Game;

import java.awt.*;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.stream.Stream;

class ArchiveList {
    private static final String SEARCHING_TEXT =
            Main.TEXT.getString("ArchiveList.searching");
    private static final String PACKAGES_FOUND_TEXT =
            Main.TEXT.getString("ArchiveList.packagesFound");

    static class PathEntry {
        final Path path;
        final UpdatePackage pkg;

        PathEntry(Path path, UpdatePackage pkg) {
            this.path = path;
            this.pkg = pkg;
        }
    }

    private HashMap<Path, UpdatePackage> foundPackages = new HashMap<>();

    boolean isEmpty() {
        return foundPackages.isEmpty();
    }
    long getPackageCount(Game game) {
        return foundPackages.values().stream().filter(pkg -> pkg.game == game).count();
    }
    UpdatePackage[] getPackages() {
        return foundPackages.values().toArray(new UpdatePackage[0]);
    }

    PathEntry[] getPaths() {
        PathEntry[] entries = foundPackages.entrySet().stream()
                .map(entry -> new PathEntry(entry.getKey(), entry.getValue()))
                .toArray(PathEntry[]::new);
        Arrays.sort(entries, Comparator.comparing(a -> a.pkg.game.shortName + " " + a.pkg.display));
        return entries;
    }

    void removePath(Path p) {
        foundPackages.remove(p);
    }

    private static Path getCanonicalPath(Path p) {
        try {
            return p.toFile().getCanonicalFile().toPath().toAbsolutePath();
        } catch (IOException e) {
            return p.toAbsolutePath();
        }
    }
    private class FindDirectoryIter {
        private int foundCount = 0;
        HashMap<Path, UpdatePackage> foundPackages = new HashMap<>();
        {
            this.foundPackages.putAll(ArchiveList.this.foundPackages);
        }

        int iter(String task, IProgressListener progress, Path path) throws IOException {
            Path absolutePath = getCanonicalPath(path);

            if(!foundPackages.containsKey(absolutePath) && Files.exists(absolutePath.resolve("update.exe"))) {
                System.out.println(" - Found potential update package in "+path+"...");
                try {
                    UpdatePackage pkg = UpdatePackage.loadPackage(path);
                    System.out.println("   - Display name  : "+pkg.display);
                    System.out.println("   - File list hash: "+pkg.fileEntryHash());
                    foundPackages.put(path, pkg);
                    foundCount++;
                    progress.setTask(task+" "+String.format(PACKAGES_FOUND_TEXT, foundCount));
                } catch(UpdaterException e) {
                    System.out.println("   - Failed to load package: "+e.getMessage());
                }
            }
            try (Stream<Path> paths = Files.list(path)) {
                for (Iterator<Path> it = paths.iterator(); it.hasNext(); ) {
                    progress.checkCancel();
                    Path child = it.next();
                    if(Files.isDirectory(child)) iter(task, progress, child);
                }
            }
            return foundCount;
        }
    }

    int findPackagesInDirectory(IProgressListener progress, Path path) throws IOException {
        progress.setCancelable(true);
        progress.setTask(SEARCHING_TEXT);

        System.out.println("Processing directory "+path+"...");
        FindDirectoryIter iter = new FindDirectoryIter();
        int count = iter.iter(SEARCHING_TEXT, progress, path);
        progress.checkCancel();
        foundPackages = iter.foundPackages;

        progress.done();
        return count;
    }
}
