package aura.cm3d2.ui;

import aura.cm3d2.core.IProgressListener;
import com.intellij.uiDesigner.core.GridConstraints;
import com.intellij.uiDesigner.core.GridLayoutManager;

import javax.swing.*;
import java.awt.*;
import java.util.ResourceBundle;

class ProgressDialog extends JDialog implements IProgressListener {
    private static final String REMAINING_STATUS = Main.TEXT.getString("ProgressDialog.remaining");
    private static final int SMOOTH_BUFFER_SIZE = 150;

    private final int MAX_PROGRESS_VALUE = 0x40000000;

    private JPanel contentPane;
    private JProgressBar progressCurrent;
    private JLabel labelCurrentStatus;
    private JButton cancelButton;
    private JScrollPane scrollPaneCurrentStatus;

    ProgressDialog() {
        setContentPane(contentPane);
        setModal(true);
        setIconImages(Main.ICONS);
        setResizable(false);

        setMinimumSize(new Dimension(550, 0));
        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);

        progressCurrent.setMaximum(MAX_PROGRESS_VALUE);
        setIndeterminate();

        cancelButton.addActionListener(event -> cancelAction());

        // Not doable in the form editor
        scrollPaneCurrentStatus.setViewportBorder(scrollPaneCurrentStatus.getBorder());
    }

    private boolean isIndeterminate = false;

    @Override
    public void setIndeterminate() {
        if (isCanceled) return;
        isIndeterminate = true;
        progressCurrent.setIndeterminate(true);
        progressCurrent.setString("");
    }

    private int historyIndex;
    private long[] bytesDoneHistory;
    private long[] readTimeHistory;

    private void setDeterminate() {
        if (!isIndeterminate) return;

        historyIndex = 0;
        bytesDoneHistory = new long[SMOOTH_BUFFER_SIZE];
        readTimeHistory = new long[SMOOTH_BUFFER_SIZE];

        long startTime = System.currentTimeMillis();
        for (int i = 0; i < readTimeHistory.length; i++) readTimeHistory[i] = startTime;

        isIndeterminate = false;
        progressCurrent.setIndeterminate(false);
    }

    // Modified from https://stackoverflow.com/a/24805871
    private static final String[] UNIT_LIST = new String[]{"KiB", "MiB", "GiB", "TiB", "PiB", "EiB"};

    private static String formatFileSize(long value) {
        int bitSize = (63 - Long.numberOfLeadingZeros(value));
        int siScale = (bitSize / 10) - 1;
        if (siScale < 0) siScale = 0;
        return String.format("%.2f %s", (double) value / (1L << ((siScale + 1) * 10)), UNIT_LIST[siScale]);
    }

    private static final String SECONDS = Main.TEXT.getString("ProgressDialog.seconds");
    private static final String MINUTES = Main.TEXT.getString("ProgressDialog.minutes");
    private static final String HOURS = Main.TEXT.getString("ProgressDialog.hours");

    public static String formatInterval(long seconds) {
        if (seconds < 60) return String.format(SECONDS, seconds);
        double minutes = (double) seconds / 60;
        if (minutes < 60) return String.format(MINUTES, Math.round(minutes));
        double hours = minutes / 60;
        return String.format(HOURS, hours);
    }

    @Override
    public void setWork(long bytesDone, long totalBytes) {
        if (isCanceled) return;
        setDeterminate();

        int index = this.historyIndex++;
        this.historyIndex %= SMOOTH_BUFFER_SIZE;

        long lastBytesDone = bytesDoneHistory[index];
        bytesDoneHistory[index] = bytesDone;

        long currentTime = System.currentTimeMillis();
        long smoothPeriodStart = readTimeHistory[index];
        readTimeHistory[index] = currentTime;

        long smoothPeriodTransferred = bytesDone - lastBytesDone;
        if (smoothPeriodTransferred < 0) smoothPeriodTransferred = 0;

        long smoothPeriodDuration = currentTime - smoothPeriodStart;
        if (smoothPeriodDuration <= 0) smoothPeriodDuration = 1;

        long remainingBytes = totalBytes - bytesDone;
        long bytesPerSecond = (smoothPeriodTransferred * 1000L) / smoothPeriodDuration;
        double progress = (double) bytesDone / (double) totalBytes;
        String barStatus = String.format(REMAINING_STATUS,
                formatFileSize(bytesDone), formatFileSize(totalBytes),
                formatInterval(remainingBytes / bytesPerSecond),
                formatFileSize(bytesPerSecond));

        progressCurrent.setValue((int) Math.round(progress * MAX_PROGRESS_VALUE));
        progressCurrent.setString(barStatus);
    }

    @Override
    public void setTask(String status) {
        if (isCanceled) return;
        labelCurrentStatus.setText(status);
    }

    @Override
    public void done() {
        dispose();
    }

    private volatile boolean isCanceled = false;

    @Override
    public void setCancelable(boolean canCancel) {
        boolean wasEnabled = cancelButton.isEnabled();
        if (canCancel != wasEnabled) {
            cancelButton.setVisible(true);
            this.pack();
            cancelButton.setEnabled(canCancel);
            isCanceled = false;
        }
    }

    @Override
    public boolean canceled() {
        return isCanceled;
    }

    private void cancelAction() {
        isCanceled = true;
        progressCurrent.setIndeterminate(true);
        progressCurrent.setString("");
        labelCurrentStatus.setText(Main.TEXT.getString("ProgressDialog.canceling"));
    }

    private volatile boolean error = false;

    boolean hasErrorOccurred() {
        return error;
    }

    @Override
    public void errorOccurred() {
        error = true;
        dispose();
    }

    {
// GUI initializer generated by IntelliJ IDEA GUI Designer
// >>> IMPORTANT!! <<<
// DO NOT EDIT OR ADD ANY CODE HERE!
        $$$setupUI$$$();
    }

    /**
     * Method generated by IntelliJ IDEA GUI Designer
     * >>> IMPORTANT!! <<<
     * DO NOT edit this method OR call it in your code!
     *
     * @noinspection ALL
     */
    private void $$$setupUI$$$() {
        contentPane = new JPanel();
        contentPane.setLayout(new GridLayoutManager(1, 1, new Insets(10, 10, 10, 10), -1, -1));
        final JPanel panel1 = new JPanel();
        panel1.setLayout(new GridLayoutManager(2, 3, new Insets(0, 0, 0, 0), -1, -1));
        contentPane.add(panel1, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null, 0, false));
        progressCurrent = new JProgressBar();
        progressCurrent.setStringPainted(true);
        progressCurrent.putClientProperty("html.disable", Boolean.TRUE);
        panel1.add(progressCurrent, new GridConstraints(0, 0, 1, 3, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_FIXED, GridConstraints.SIZEPOLICY_FIXED, null, null, null, 0, false));
        cancelButton = new JButton();
        cancelButton.setEnabled(false);
        this.$$$loadButtonText$$$(cancelButton, ResourceBundle.getBundle("aura/cm3d2/ui/text").getString("ProgressDialog.cancel"));
        cancelButton.setToolTipText(ResourceBundle.getBundle("aura/cm3d2/ui/text").getString("ProgressDialog.cancel.tooltip"));
        cancelButton.setVisible(false);
        cancelButton.putClientProperty("html.disable", Boolean.TRUE);
        panel1.add(cancelButton, new GridConstraints(1, 2, 1, 1, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null, 0, false));
        scrollPaneCurrentStatus = new JScrollPane();
        scrollPaneCurrentStatus.setEnabled(false);
        scrollPaneCurrentStatus.setHorizontalScrollBarPolicy(31);
        scrollPaneCurrentStatus.setVerticalScrollBarPolicy(21);
        panel1.add(scrollPaneCurrentStatus, new GridConstraints(1, 0, 1, 2, GridConstraints.ANCHOR_CENTER, GridConstraints.FILL_BOTH, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, GridConstraints.SIZEPOLICY_CAN_SHRINK | GridConstraints.SIZEPOLICY_WANT_GROW, null, null, null, 0, false));
        scrollPaneCurrentStatus.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEmptyBorder(), null));
        final JPanel panel2 = new JPanel();
        panel2.setLayout(new GridLayoutManager(1, 1, new Insets(0, 0, 0, 0), -1, -1));
        scrollPaneCurrentStatus.setViewportView(panel2);
        panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEmptyBorder(), null));
        labelCurrentStatus = new JLabel();
        labelCurrentStatus.setEnabled(true);
        labelCurrentStatus.setText("[Current Status]");
        labelCurrentStatus.putClientProperty("html.disable", Boolean.TRUE);
        panel2.add(labelCurrentStatus, new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_NONE, 1, GridConstraints.SIZEPOLICY_FIXED, null, null, null, 0, false));
    }

    /**
     * @noinspection ALL
     */
    private void $$$loadButtonText$$$(AbstractButton component, String text) {
        StringBuffer result = new StringBuffer();
        boolean haveMnemonic = false;
        char mnemonic = '\0';
        int mnemonicIndex = -1;
        for (int i = 0; i < text.length(); i++) {
            if (text.charAt(i) == '&') {
                i++;
                if (i == text.length()) break;
                if (!haveMnemonic && text.charAt(i) != '&') {
                    haveMnemonic = true;
                    mnemonic = text.charAt(i);
                    mnemonicIndex = result.length();
                }
            }
            result.append(text.charAt(i));
        }
        component.setText(result.toString());
        if (haveMnemonic) {
            component.setMnemonic(mnemonic);
            component.setDisplayedMnemonicIndex(mnemonicIndex);
        }
    }

    /**
     * @noinspection ALL
     */
    public JComponent $$$getRootComponent$$$() {
        return contentPane;
    }
}
